@@ -29,6 +29,74 @@ def exponential_linspace_int(start, end, num, divisible_by = 1):
     base = math.exp(math.log(end / start) / (num - 1))
     return [_round(start * base**i) for i in range(num)]
 
+# relative positional encoding functions
+
+def get_positional_features_exponential(positions, features, seq_len, min_half_life = 3.):
+    max_range = math.log(seq_len) / math.log(2.)
+    half_life = 2 ** torch.linspace(min_half_life, max_range, features, device = positions.device)
+    half_life = half_life[None, ...]
+    positions = positions.abs()[..., None]
+    return torch.exp(-math.log(2.) / half_life * positions)
+
+def get_positional_features_central_mask(positions, features, seq_len):
+    center_widths = 2 ** torch.arange(1, features + 1, device = positions.device).float()
+    center_widths = center_widths - 1
+    return (center_widths[None, ...] > positions.abs()[..., None]).float()
+
+def gamma_pdf(x, concentration, rate):
+    log_unnormalized_prob = torch.xlogy(concentration - 1., x) - rate * x
+    log_normalization = (torch.lgamma(concentration) - concentration * torch.log(rate))
+    return torch.exp(log_unnormalized_prob - log_normalization)
+
+def get_positional_features_gamma(positions, features, seq_len, stddev = None, start_mean = None, eps = 1e-8):
+    if not exists(stddev):
+        stddev = seq_len / (2 * features)
+
+    if not exists(start_mean):
+        start_mean = seq_len / features
+
+    mean = torch.linspace(start_mean, seq_len, features)
+    mean = mean[None, ...]
+    concentration = (mean / stddev) ** 2
+    rate = mean / stddev ** 2
+    probabilities = gamma_pdf(positions.float().abs()[..., None], concentration, rate)
+    probabilities = probabilities + eps
+    outputs = probabilities / torch.amax(probabilities)
+    return outputs
+
+def get_positional_embed(seq_len, feature_size, device):
+    distances = torch.arange(-seq_len + 1, seq_len, device = device)
+
+    feature_functions = [
+        get_positional_features_exponential,
+        get_positional_features_central_mask,
+        get_positional_features_gamma
+    ]
+
+    num_components = len(feature_functions) * 2
+
+    if (feature_size % num_components) != 0:
+        raise ValueError(f'feature size is not divisible by number of components ({num_components})')
+
+    num_basis_per_class = feature_size // num_components
+
+    embeddings = []
+    for fn in feature_functions:
+        embeddings.append(fn(distances, num_basis_per_class, seq_len))
+
+    embeddings = torch.cat(embeddings, dim = -1)
+    embeddings = torch.cat((embeddings, torch.sign(distances)[..., None] * embeddings), dim = -1)
+    return embeddings
+
+def relative_shift(x):
+    to_pad = torch.zeros_like(x[..., :1])
+    x = torch.cat((to_pad, x), dim = -1)
+    _, h, t1, t2 = x.shape
+    x = x.reshape(-1, h, t2, t1)
+    x = x[:, :, 1:, :]
+    x = x.reshape(-1, h, t1, t2 - 1)
+    return x[..., :((t2 + 1) // 2)]
+
 # classes
 
 class Residual(nn.Module):
@@ -87,10 +155,12 @@ class Attention(nn.Module):
         self,
         dim,
         *,
+        num_rel_pos_features,
         heads = 8,
         dim_key = 64,
         dim_value = 64,
-        dropout = 0.
+        dropout = 0.,
+        pos_dropout = 0.
     ):
         super().__init__()
         self.scale = dim_key ** -0.5
@@ -99,20 +169,46 @@ class Attention(nn.Module):
         self.to_q = nn.Linear(dim, dim_key * heads, bias = False)
         self.to_k = nn.Linear(dim, dim_key * heads, bias = False)
         self.to_v = nn.Linear(dim, dim_value * heads, bias = False)
-        self.attn_dropout = nn.Dropout(dropout)
 
         self.to_out = nn.Linear(dim_value * heads, dim)
+        nn.init.zeros_(self.to_out.weight)
+        nn.init.zeros_(self.to_out.bias)
+
+        # relative positional encoding
+
+        self.num_rel_pos_features = num_rel_pos_features
+
+        self.to_rel_k = nn.Linear(num_rel_pos_features, dim_key * heads, bias = False)
+        self.rel_content_bias = nn.Parameter(torch.randn(1, heads, 1, dim_key))
+        self.rel_pos_bias = nn.Parameter(torch.randn(1, heads, 1, dim_key))
+
+        # dropouts
+
+        self.pos_dropout = nn.Dropout(pos_dropout)
+        self.attn_dropout = nn.Dropout(dropout)
 
     def forward(self, x):
-        h = self.heads
+        n, h, device = x.shape[-2], self.heads, x.device
+
         q = self.to_q(x)
         k = self.to_k(x)
         v = self.to_v(x)
 
         q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
 
-        sim = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
-        attn = sim.softmax(dim = -1)
+        q = q * self.scale
+
+        content_logits = einsum('b h i d, b h j d -> b h i j', q + self.rel_content_bias, k)
+
+        positions = get_positional_embed(n, self.num_rel_pos_features, device)
+        rel_k = self.to_rel_k(positions)
+
+        rel_k = rearrange(rel_k, 'n (h d) -> h n d', h = h)
+        rel_logits = einsum('b h i d, h j d -> b h i j', q + self.rel_pos_bias, rel_k)
+        rel_logits = relative_shift(rel_logits)
+
+        logits = content_logits + rel_logits
+        attn = logits.softmax(dim = -1)
         attn = self.attn_dropout(attn)
 
         out = einsum('b h i j, b h j d -> b h i d', attn, v)
@@ -132,9 +228,12 @@ class Enformer(nn.Module):
         target_length = TARGET_LENGTH,
         dropout_rate = 0.4,
         num_alphabet = 5,
-        attn_dim_key = 64
+        attn_dim_key = 64,
+        attn_dropout = 0.05,
+        pos_dropout = 0.01
     ):
         super().__init__()
+        self.num_alphabet = num_alphabet
         half_dim = dim // 2
         twice_dim = dim * 2
 
@@ -173,7 +272,10 @@ class Enformer(nn.Module):
                         dim,
                         heads = heads,
                         dim_key = attn_dim_key,
-                        dim_value = dim // heads
+                        dim_value = dim // heads,
+                        dropout = attn_dropout,
+                        pos_dropout = pos_dropout,
+                        num_rel_pos_features = dim // heads
                     ),
                     nn.Dropout(dropout_rate)
                 )),
@@ -231,5 +333,8 @@ class Enformer(nn.Module):
         return self._heads
     
     def forward(self, x):
+        if x.dtype == torch.long:
+            x = F.one_hot(x, num_classes = self.num_alphabet)
+
         x = self._trunk(x.float())
         return map_values(lambda fn: fn(x), self._heads)
